/*
 * fpbus_ng.c
 *
 * Copyright (C) 2006 Luotao Fu <lfu@pengutronix.de>, Pengutronix
 *                    
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the version 2 of the GNU General Public License
 * as published by the Free Software Foundation
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* fpga pin setup:						*/
/* /config         UART6_CTS_TTL   PSC6_1  GPIO_WKUP_5     out	*/
/* config_done     UART6_RXD_TTL   PSC6_0  GPIO_WKUP_4     in	*/
/* /status         GPIO7                   GPIO_WKUP_7     in	*/
/* dclk            UART6_RTS_TTL   PSC6_3  GPIO_IRDA_1     out	*/
/* data0           UART6_TXD_TTL   PSC6_2  GPIO_IRDA_0     out	*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/delay.h> 

#include <linux/firmware.h>
#include <linux/errno.h> 
#include <linux/types.h>
#include <linux/fcntl.h> 
#include <linux/fs.h> 

#include <asm/bootinfo.h>
#include <asm/io.h>
#include <asm/mpc52xx.h>
#include <asm/ppc_sys.h>
#include <asm/machdep.h>

#include <syslib/mpc52xx_pci.h>

#define TIMEOUT 50000
#define DRIVERNAME "fpga"

static char *firmware = NULL;
module_param(firmware, charp, 0400);
MODULE_PARM_DESC(firmware, "firmware filename");


static void fpga_dev_release (struct device *dev) {
}


static struct platform_device fpga_device = {
	.dev.release	= fpga_dev_release,
        .name           = DRIVERNAME,
        .id             = 0,
};


static int fpga_drv_probe (struct device *dev) {
	struct mpc52xx_gpio __iomem *gpio;
	struct mpc52xx_gpio_wkup __iomem *gpio_wkup;
	int bits_transferred = 0, size, timeout = TIMEOUT, err;
	const struct firmware *fw;
	u32 simple_gpioe, simple_ddr, port_config;
	u8 wkup_gpioe, wkup_ddr;
	u8 config, conf_done, status;
	u32 data;
	u8 *buf;

	err = request_firmware(&fw, firmware, dev);
	if (err){
		dev_err(dev, "%s: could not load firmware!%d\n", __FUNCTION__, err);
		return err;
	}

	buf = fw->data;
	size = fw->size;

	gpio = ioremap(MPC52xx_PA(MPC52xx_GPIO_OFFSET), MPC52xx_GPIO_SIZE);
	gpio_wkup = ioremap(MPC52xx_PA(MPC52xx_GPIO_WKUP_OFFSET), 
			    MPC52xx_GPIO_WKUP_SIZE);
        if (!gpio || !gpio_wkup) {
		printk("%s: Error while mapping GPIO/GPIO_WKUP", __FUNCTION__);
		goto unmap_regs;
	}

	/*set Port to configure IRDA as GPIO*/
	port_config = in_be32(&gpio->port_config);
	port_config &= ~0x00700000;
	out_be32(&gpio->port_config, port_config);

	/* enabling GPIO_IRDA_0, GPIO_IRDA_1 */
	simple_gpioe = in_be32(&gpio->simple_gpioe);
	simple_gpioe |= 0x30000000;
	out_be32(&gpio->simple_gpioe, simple_gpioe);

	/* set GPIO_IRDA_0 as out(Bit 2 low), GPIO_IRDA_1 as out(Bit 3 high)*/
	simple_ddr = in_be32(&gpio->simple_ddr);
	simple_ddr |= 0x30000000;
	out_be32(&gpio->simple_ddr, simple_ddr);
		
	/* enabling GPIO_WKUP_4, GPIO_WKUP_5, GPIO_WKUP_7 */
	wkup_gpioe = in_8(&gpio_wkup->wkup_gpioe);
	wkup_gpioe |= 0xB0;
	out_8(&gpio_wkup->wkup_gpioe, wkup_gpioe);	

	/* set GPIO_WKUP_7 GPIO_4 as in (Bit 0 and Bit 3 low)and GPIO_WKUP_5 */
	/* as out (Bit 2 high) */
	wkup_ddr = in_8(&gpio_wkup->wkup_ddr);
	wkup_ddr &= ~0x90;
	wkup_ddr |= 0x20;
	out_8(&gpio_wkup->wkup_ddr, wkup_ddr);

	/* set config(GPIO_WKUP_5) to low and wait 50 us till the  */
	/* initialization of the datatransfer */
	config = in_8(&gpio_wkup->wkup_dvo);
	config &= ~0x20;
	out_8(&gpio_wkup->wkup_dvo, config);
	udelay(50);

	/*set config to high and wait till device ready (high on status)*/
	config = in_8(&gpio_wkup->wkup_dvo);
	config |= 0x20;
	out_8(&gpio_wkup->wkup_dvo, config);

	while (1) {
		status = in_8(&gpio_wkup->wkup_ival);
		status &= 0x80;
		if (status)
			break;
	        if (timeout <= 0)
			return -ENODEV;
		timeout--;
		udelay(1);
	}

	printk("flashing firmware %dkbyte to go.", size);

	while (1) {
		conf_done = in_8(&gpio_wkup->wkup_ival);
		conf_done &= 0x10;

		if (conf_done)
			break;		

		/* printk("."); */

		if (*buf & 0x1){
 			/*set data pin IRDA_0*/ 
			data = in_be32(&gpio->simple_dvo); 
			data |= 0x10000000;
			out_be32(&gpio->simple_dvo, data);	
		} else {
			/*clear data pin*/
			data = in_be32(&gpio->simple_dvo); 
			data &= ~0x10000000;
			out_be32(&gpio->simple_dvo, data);
		}

		/*set dclk*/
		data = in_be32(&gpio->simple_dvo); 
		data |= 0x20000000;
		out_be32(&gpio->simple_dvo, data);		

		/*clear dclk*/
		data = in_be32(&gpio->simple_dvo); 
		data &= ~0x20000000;

		out_be32(&gpio->simple_dvo, data);	
		*buf >>= 1;

		if (!(++bits_transferred % 8))
			buf++;

 		if(!(bits_transferred % 8192))
			printk(".");

		if (bits_transferred > size * 8){
			printk("Warning: Bit overrun\n");
			break;
		}
	}
	printk("done\n");
	printk("exit value: %x %x %d\n", conf_done, 
	       in_8(&gpio_wkup->wkup_ival), bits_transferred);

unmap_regs:
	if (gpio) 
		iounmap(gpio);
	if (gpio_wkup) 
		iounmap(gpio_wkup);

	release_firmware(fw);	
	return 0;
}

static int fpga_drv_remove (struct device *dev) {
	return 0;
}

static struct device_driver fpga_driver = {
	.name		= DRIVERNAME,
	.bus		= &platform_bus_type,
	.probe		= fpga_drv_probe,
	.remove		= fpga_drv_remove,
#ifdef CONFIG_PM
	.suspend	= fpga_drv_suspend,
	.resume		= fpga_drv_resume,
#endif
};

static int __init fpga_init(void) {	
	int err;

	if (!firmware) {
		printk(KERN_ERR "No firmware file specified!\n");
		return -EINVAL;	
	}      

	err = platform_device_register(&fpga_device);
	if (err) {
		printk(KERN_ERR "cannot register device\n");
		goto exit;
	}

	err = driver_register(&fpga_driver);
        if (err) {
		printk(KERN_ERR "cannot register driver\n");
		goto exit_unregister_device;	
	}

	return 0;

 exit_unregister_device:
	platform_device_unregister(&fpga_device);
 exit:
	return err;
}

static void __exit fpga_cleanup(void) {
	driver_unregister(&fpga_driver);
	platform_device_unregister(&fpga_device);
}

module_init(fpga_init);
module_exit(fpga_cleanup);
MODULE_AUTHOR("Pengutronix lfu");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Driver for FPGA on a MPC52XXX Board");
